#!/usr/bin/lua

local ffi = require 'ffi'
local C = ffi.C
local lpeg = require 'lpeg'
local bit = require 'bit'
local event = ...

ffi.cdef[[
char *getenv(const char *name);
int unlink(const char *pathname);
int rename(const char *oldpath, const char *newpath);
]]

local params = { }
local started = false
local plist = {
    'interface',
    'subnet',
    'broadcast',
    'ip',
    'router',
    'domain',
    'dns'}

for _, key in ipairs(plist) do
    local v = C.getenv(key)
    if v ~= nil then
        params[key] = ffi.string(v)
    end
end

local function split(s, sep)
    sep = lpeg.P(sep)
    local elem = lpeg.C((1 - sep)^0)
    local p = lpeg.Ct(elem * (sep * elem)^0)
    return lpeg.match(p, s)
end

local function mask_to_bits(ip, mask)
    local bits = ip:match('^.+/(.+)$')
    if bits and tonumber(bits) then return tonumber(bits) end
    bits = 0
    local c = {mask:match('^(%d-)%.(%d-)%.(%d-)%.(%d+)$')}
    if #c == 4 then
        local mask = 0
        for _, v in ipairs(c) do
            for i = 0, 7 do
                if bit.band(v, 0x80) == 0x80 then bits = bits + 1 end
                v = bit.lshift(v, 1)
            end
        end
    end
    return bits
end

local configs = {}

-- This bit of code was taken from network.lua
local p, e = params, os.execute
if event == 'deconfig' then
    if configs[p.interface] then
        e('/sbin/ip addr del ' .. configs[p.interface] ..
            ' dev ' ..  p.interface)
        configs[p.interface] = nil
    end
elseif event == 'renew' or event == 'bound' then
    local bits = mask_to_bits(p.ip, p.subnet)
    local broadcast = p.broadcast and 'broadcast ' .. p.broadcast or
        'broadcast +'
    e('/sbin/ip addr add ' .. p.ip .. '/' .. tostring(bits) .. ' ' ..
        broadcast .. ' dev ' .. p.interface)
    configs[p.interface] = p.ip .. '/' .. tostring(bits)
    if p.router then
        while e('/sbin/route del default gw 0.0.0.0 dev ' .. p.interface ..
            ' 2>/dev/null') == 0 do end
        local metric = 0
        for _, router in ipairs(split(p.router, ' ')) do
            e('/sbin/route add default gw ' .. router .. ' dev ' ..
                p.interface .. '  metric ' .. metric)
            metric = metric + 1
        end
    end

    os.execute('mkdir -p /etc/resolv.conf.d')
    local rc = io.open('/etc/resolv.conf.d/resolv.conf.' .. p.interface, 'w')
    local nss = {}
    for _, ns in ipairs(split(p.dns, ' ')) do
        rc:write('nameserver ' .. ns .. '\n')
    end
    if p.domain then
        rc:write('search ' .. p.domain .. '\n')
    end
    rc:close()
    os.execute('/usr/bin/combine-resolv-conf')
end
