--! \file

local common = require 'mx.device_common'
local underscore = require 'underscore'
local ffi = require 'ffi'
local C = ffi.C
local m = {}

-- XML used to do initial configuration of channels when proxy starts up
local default_xml = [[<?xml version="1.0" encoding="UTF-8"?>
                <Amplifier>
                    <Connector#conn_num#>
                        <Mode choice="AnalogInChannel">
                            <AnalogInChannel>
                                <Sensor>
                                    <TedsOptions>
                                        <UseTeds>2</UseTeds>
                                    </TedsOptions>
                                </Sensor>
                                <SyncSignal address="1">
                                    <OutputRate>10</OutputRate>
                                </SyncSignal>
                            </AnalogInChannel>
                        </Mode>
                    </Connector#conn_num#>
                </Amplifier>
]]

-- Function used to return the XML used to set the channel to it's default
-- configuration
function m:get_default_xml(channel_num)
    return default_xml:gsub('#conn_num#', tostring(channel_num))
end

m.json2xpath = {
    transducer_power = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Supply',
    _s2enbl = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/S2Enbl',
    _f4enbl = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/F4Enbl',
    _indexenbl = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/IndexEnbl',
    _indexdivisor = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/IndexDivisor',

    special_parameters = {
        zero_shift = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/SpecialParameters/IndexShift',
        interpolation_enable = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/SpecialParameters/InterpolationEnbl',
        glitch_filter = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/SpecialParameters/GlitchFilter',
        invert_polarity = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/SpecialParameters/Polarity',
        termination = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/SpecialParameters/Termination',
    },

    internal_termination = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/SpecialParameters/Termination',

    sensor_input_extra = {
        pulses_per_round = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/PulsePerRound',
        gap_position = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/GapPosition',
        pulse_edge = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/CountingEdge',
        zero_position = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/IndexPosition',
        zero_edge = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/IndexEdge',
--        zero_divisor = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/IndexDivisor',
   },
}

underscore.extend(m.json2xpath, common.json2xpath)

m.phase_delay = {}
m.phase_delay.fixed = 398
m.phase_delay.bessel_iir = {}
m.phase_delay.butterworth_iir = {}
m.phase_delay.linear_phase_fir = {}
m.phase_delay.butterworth_fir = {}

m.physical_unit_map = {
    Frequency = 'kHz',
    Counter = 'kImp',
    Pwm = 'ms',
    IndCounter = 'kImp',
    IndFrequency = 'kHz',
    PulseWidth = 'ms',
    Period = 'ms',
    IndCrankWheelCounter = 'kImp',
    IndCrankWheelFrequency = 'kHz',
    CrankWheelCounter = 'kImp',
    CrankWheelFrequency = 'kHz',
}

-- List of parameters in a channel setup that we do NOT want sent off to the MX
-- device
m.ignore_list = {}
underscore.extend(m.ignore_list, common.ignore_list)

-- List of parameters returned by MX device that we don't verify after setting
-- up a channel.
m.dont_verify_list = {}
underscore.extend(m.dont_verify_list, common.dont_verify_list)

function m:internal_termination(p, fromxml)
    if fromxml then
        return tonumber(p) == 1
    end
    return p and 1 or 0
end

function m:invert_polarity(p, fromxml)
    if fromxml then
        return tonumber(p) == 1
    end
    return p and 1 or 0
end

function m:gap_position(p, fromxml)
    if fromxml then return "none" end
    if type(p) == 'string' and p:lower() == 'none' then return '' end
    return p
end

m.translator = {
    internal_termination = m.internal_termination,
    invert_polarity = m.invert_polarity,
    gap_position = m.gap_position,
}

function m:to_internal(spec)
    local spec = spec
    -- Do some processing based on sensor input mode (if it exists)
    if spec.sensor_input and spec.sensor_input.mode then
        if spec.sensor_input.mode == 'pulse_counter' then
            spec._s2enbl = '0'
            spec._f4enbl = '0'
            spec._indexenbl = '0'
            spec._indexdivisor = '1'
        elseif spec.sensor_input.mode == 'encoder_counter' then
            spec._s2enbl = '1'
            if spec.sensor_input.track_sign ~= nil then
                spec._s2enbl = spec.sensor_input.track_sign and '1' or '0'
            end
            spec._f4enbl = spec.sensor_input.quadrature and '1' or '0'
            spec._indexenbl =
                tostring(spec.sensor_input.encoder_index_reset or 'false') ==
                    'true' and '1' or '0'
            spec._indexdivisor =
                tostring(spec.sensor_input.encoder_index_divisor or '1')
        elseif spec.sensor_input.mode == 'pulse_frequency' then
            spec._s2enbl = '0'
            spec._f4enbl = '0'
        elseif spec.sensor_input.mode == 'encoder_frequency' then
            spec._s2enbl = '1'
            if spec.sensor_input.track_sign ~= nil then
                spec._s2enbl = spec.sensor_input.track_sign and '1' or '0'
            end
            spec._f4enbl = spec.sensor_input.quadrature and '1' or '0'
        end
    end

    if spec.internal_termination and spec.special_parameters then
        spec.special_parameters.termination =
            m.internal_termination(self, spec.internal_termination)
        spec.internal_termination = nil
    end

    if spec.special_parameters then
        if spec.special_parameters.invert_polarity then
            spec.special_parameters.invert_polarity = m.invert_polarity(
                self, spec.special_parameters.invert_polarity)
        end
    end

    if spec.sensor_input and spec.sensor_input.gap_position then
        spec.sensor_input.gap_position =
            m.gap_position(self, spec.sensor_input.gap_position)
    end

    return spec
end

local zero_xml_template =[[
<?xml version="1.0"?>
<SrvParams>
<Control>SETZERO</Control>
<ControlParams>&lt;?xml version="1.0"
encoding="UTF-8"?&gt;&lt;ControlParams&gt;&lt;Par1&gt;#channel&lt;/Par1&gt;&lt;Par2&gt;&lt;/Par2&gt;&lt;Par3&gt;#target&lt;/Par3&gt;&lt;/ControlParams&gt;</ControlParams>
</SrvParams>
]]

m.controls = {}
m.controls.channel = {}

-- Function used to control channel zeroing.  This is really only used for
-- interactive zeroing of counter sensor types.
-- Note, self in this function is the ChannelProxy object
function m.controls.channel:zero(params, cb)
    local chan = self.index
    local target = params and params[1]
    local xml = zero_xml_template
    xml = string.gsub(xml, '#channel', chan)
    xml = string.gsub(xml, '#target', target or 0.0)
    self.mxdev.mx.control:_send_transaction_request(C.TR_SCHEMA, C.TR_XML,
        function(obj, ecode, data, len)
            cb{result = 'success', msg = 'ok'}
        end, xml)
end

function m:to_external(spec)
    local spec = spec
    spec.calibration.units_conversion = nil

    if spec.sensor_input.mode and (spec.sensor_input.mode == 'pulse_counter' or
            spec.sensor_input.mode == 'encoder_counter') then
        if spec._s2enbl == '0' then
            spec.sensor_input.mode = 'pulse_counter'
        else
            spec.sensor_input.mode = 'encoder_counter'
            spec.sensor_input.encoder_index_reset =
                spec._indexenbl == '1' and true or false
            if spec.sensor_input.encoder_index_reset then
                spec.sensor_input.encoder_index_divisor =
                    tonumber(spec._indexdivisor or 1)
            end
        end
        if spec.sensor_input.mode == 'encoder_counter' then
            spec.sensor_input.quadrature = spec._f4enbl == '1'
            spec.sensor_input.track_sign = spec._s2enbl == '1'
        end
    end

    if spec.sensor_input.mode and
        (spec.sensor_input.mode == 'pulse_frequency' or
        spec.sensor_input.mode == 'encoder_frequency') then
        if spec._s2enbl == '0' then
            spec.sensor_input.mode = 'pulse_frequency'
        else
            spec.sensor_input.mode = 'encoder_frequency'
        end
        if spec.sensor_input.mode == 'encoder_frequency' then
            spec.sensor_input.quadrature = spec._f4enbl == '1'
            spec.sensor_input.track_sign = spec._s2enbl == '1'
        end
    end

    for k, v in pairs(spec.sensor_input_extra) do
        spec.sensor_input[k] = tonumber(v) or v
    end
    spec.sensor_input_extra = nil

    spec._s2enbl = nil
    spec._f4enbl = nil
    spec._indexenbl = nil
    spec._indexdivisor = nil

    -- Based on the input mode limit the special parameters to only those that
    -- apply
    local param_sets = {
        pulse_frequency = {
            glitch_filter = true,
            invert_polarity = true,
            termination = true,
        },
        encoder_frequency = {
            glitch_filter = true,
            invert_polarity = true,
            termination = true,
        },
        pulse_counter = {
            zero_shift = true,
            interpolation_enable = true,
            glitch_filter = true,
            invert_polarity = true,
            termination = true,
        },
        encoder_counter = {
            zero_shift = true,
            interpolation_enable = true,
            glitch_filter = true,
            invert_polarity = true,
            termination = true,
        },
        inductive_pulse_frequency = {
            glitch_filter = true,
        },
        inductive_pulse_counter = {
            interpolation_enable = true,
            glitch_filter = true,
            invert_polarity = true,
        },
        pulse_period = {
            glitch_filter = true,
            invert_polarity = true,
            termination = true,
        },
        pulse_high_period = {
            glitch_filter = true,
            invert_polarity = true,
            termination = true,
        },
        pulse_duty_cycle = {
            glitch_filter = true,
            invert_polarity = true,
            termination = true,
        },
        crank_wheel_frequency = {
            glitch_filter = true,
        },
        crank_wheel_counter = {
            zero_shift = true,
            interpolation_enable = true,
            glitch_filter = true,
        },
        inductive_crank_wheel_frequency = {
            glitch_filter = true,
        },
        inductive_crank_wheel_counter = {
            zero_shift = true,
            interpolation_enable = true,
            glitch_filter = true,
        },
    }

    local pset = param_sets[spec.sensor_input.mode] or {}
    for k, v in pairs(spec.special_parameters) do
        if pset[k] ~= nil then
            spec.special_parameters[k] = tonumber(v) or v
        else
            spec.special_parameters[k] = nil
        end
    end
    if spec.special_parameters.invert_polarity then
        spec.special_parameters.invert_polarity = m.invert_polarity(
            self, spec.special_parameters.invert_polarity, true)
    end

    if not spec.special_parameters.termination then
        spec.internal_termination = nil
    else
        spec.internal_termination =
            tonumber(spec.special_parameters.termination) or
            spec.special_parameters.termination
    end
    spec.special_parameters.termination = nil
    if spec.internal_termination then
        spec.internal_termination =
            m.internal_termination(self, spec.internal_termination, true)
    end

    if spec.sensor_input.gap_position then
        spec.sensor_input.gap_position =
            m.gap_position(self, spec.sensor_input.gap_position, true)
    end

    return spec
end

-- Specification of the default settings of this channel type
m.default_spec = {
    _s2enbl = '0',
    sample_rate = '10',
    sensor_input = {
        mode = 'pulse_frequency',
        min = -200.0,
        max = 200.0,
    },
    special_parameters = {
        quadrature = false,
        track_sign = false,
        termination = '0',
    },
    calibration = {
        define_source_units = {
            units = 'kHz'
        },
        sensor_scaling = {
            mode = 'defined_slope_intercept',
            slope = 1,
            intercept = 0,
            units = 'kHz'
        },
        rezero = {
            mode = 'defined_slope_intercept',
            slope = 1,
            intercept = 0
        },
    },
    digital_filter_type = 'none',
    transducer_power = 0,
}

return m
