--! \file

local dbg = require "dbg"
local lpeg = require "lpeg"
local bit = require "bit"
local membuf = require "membuf"
local gpschar = require "gpschar"

local bxor = bit.bxor
local string_byte = string.byte
local string_find = string.find
local string_len = string.len
local string_sub = string.sub
local table_insert = table.insert
local table_remove = table.remove

local log = require 'log'

local class = require "class"

local ffi = require 'ffi'
local C = ffi.C


ffi.cdef[[
    static const size_t  MAX_RE_ERROR_STR_LEN = 200;
    int isxdigit(int c);
]]


--! An NMEA parser.
--? class nmea_parser
local m = class()
--? alias m nmea_parser

function m:init()
    self.stream_pos = 0
    self.stamps = {}
    self.buf = membuf:new(nil, 0)
end

-- This function finds the ".0" suffix in the given membuf object.
-- It returns the offset to the "." character or nil if not found.
--
-- Parameter 'in_buf' is the membuf object containing the string to
--     check.
-- Parameter 'buf_end_offset' is the offset of the newline at the
--     end of the string in 'in_buf'.  This allows quicker checking
--     than starting at the front of the string.
local function find_star(in_buf, buf_end_offset)
    local star_offset = nil
    local possible_star = in_buf.buf[buf_end_offset - 4]

    if (possible_star == gpschar.CHAR_STAR) and
       (C.isxdigit(in_buf.buf[buf_end_offset - 3]) ~= 0) and
       (C.isxdigit(in_buf.buf[buf_end_offset - 2]) ~= 0) then

        star_offset = buf_end_offset - 4;
    end

    return star_offset;
end

-- This function converts a (sub)string of ASCII hex bytes to a
-- non-negative binary number.  It returns the number, or -1 if
-- an error is found.  An error can be caused by an ASCII non-hex
-- byte within the input (sub)string.  However, in the case of the
-- parameter 'hex_len' being <= zero an ASCII non-hex byte will
-- merely indicate the end of the input (sub)string to be converted.
--
-- Parameter 'in_buf' is the membuf object containing the string to
--     be converted.  'in_buf' is not modified by this function.
-- Parameter 'hex_start' is the offset from the head of 'in_buf' at
--     which the string to be converted is found.
-- Parameter 'hex_len' is the length in bytes of the string to
--     be converted.  If 'hex_len' has a value <= zero, the length
--     is made equal to the remaining bytes in 'in_buf' after
--     'hex_start'.
function ascii_hex_to_number(in_buf, hex_start, hex_len)
    local returned_nbr = 0
    local len_to_convert = 0
    local len_modified = false

    if not hex_len or hex_len <= 0 then
        len_to_convert = #in_buf - hex_start
        if len_to_convert <= 0 then
            return 0
        end
        len_modified = true
    else
        len_to_convert = hex_len
    end

    for i = hex_start, hex_start + len_to_convert - 1, 1 do
	local byte_nbr = 0;
        if (in_buf.buf[i] >= gpschar.CHAR_ZERO and
            in_buf.buf[i] <= gpschar.CHAR_NINE) then
            byte_nbr = byte_nbr + in_buf.buf[i] - gpschar.CHAR_ZERO
        elseif (in_buf.buf[i] >= gpschar.CHAR_A and
                in_buf.buf[i] <= gpschar.CHAR_F) then
            byte_nbr = byte_nbr + in_buf.buf[i] - gpschar.CHAR_A + 10
        elseif (in_buf.buf[i] >= gpschar.CHAR_a and
                in_buf.buf[i] <= gpschar.CHAR_f) then
            byte_nbr = byte_nbr + in_buf.buf[i] - gpschar.CHAR_a + 10
        else
            if not len_modified then
                returned_nbr = -1  -- error value returned!
            end
            break;
        end
        returned_nbr = returned_nbr * 16 + byte_nbr
    end

    return returned_nbr
end

local function discard(self, trim_bytes, err_text)
    if err_text then
        log:warn(string.format("discarded GPS text \"%s\" due to:  %s",
            ffi.string(self.buf:head(), trim_bytes),  err_text))
    end
--dbg.debugf("DISCARD %d\n", trim_bytes)
    local stream_pos = self.stream_pos

    self.buf:remove(0, trim_bytes)
    self.stream_pos = self.stream_pos + trim_bytes
--dbg.debugf("   new stream_pos=%d\n", self.stream_pos)

    for i = #self.stamps, 2, -1 do
        if self.stamps[i][1] <= self.stream_pos then
--dbg.debugf("   deleting %d stamps\n", i - 1)
            for j = 1, i - 1 do
                table_remove(self.stamps, 1)
            end
            break
        end
    end
end

function m:nmea_filter(text)
    local msg, csum = string.match(text, '^%$(.+)%*(%x%x).-$')
    if not msg or not csum then
        log:warn(string.format("discarded GPS text \"%s\" due to:  %s",
            text,  'Unable to parse NMEA sentence'))
        return
    end
    csum = tonumber(csum, 16)
    local sum = 0
    for i = 1, msg:len() do
        sum = bxor(sum, msg:byte(i))
    end
    if tonumber(sum) == csum then return text end
    log:warn(string.format(
        "Discarded GPS text \"$%s\" due to checksum mismatch (%02X <> %02X)",
        msg, bit.band(tonumber(sum), 0xff), bit.band(csum, 0xff)))
end

function m:nmea_filter_old(pkt)
    if pkt.length == 0 then
        return
    end
    self.stamps[#self.stamps + 1]
        = {self.stream_pos + #self.buf + 1, pkt.timestamp}

--[==[
    dbg.debugf("PKT[%g / %g]: (%s)\n", pkt.timestamp, self.stamps[#self.stamps][1], pkt.text)
for i = 1, #self.stamps do
    --dbg.debugf("    STAMPS[%d]: {%g, %g}\n", i, self.stamps[i][1], self.stamps[i][2])
end
--]==]

    self.buf = self.buf:append(pkt.text, pkt.length)
    local out = {}
    while #self.buf > 0 do
--dbg.debugf("nmea_filter loop, len=%d, buf=%s\n", string.len(self.buf), self.buf)
        local dollar_char_offset = self.buf:search(0, "$", 1)
        if dollar_char_offset ~= 0 then
            -- KLUDGE Warn about malformed stuff
            if dollar_char_offset then
                discard(self, dollar_char_offset, "invalid data before $")
            else
                self.stream_pos = self.stream_pos + #self.buf
                self.buf:clear()
                self.stamps = {}
                break
            end
        end
        -- Now self.buf starts with a "$"
        local end_offset = self.buf:search(0, "\n", 1)
        if not end_offset then
            self.buf:clear()
            break
        end

        local star_offset = find_star(self.buf, end_offset)
        if star_offset == nil then
            -- delete the current "line"
            discard(self, end_offset + 1, "sentence without terminating *")
            break
        end

        local sum = 0
        for i = 1, star_offset - 1, 1 do
            sum = bxor(sum, self.buf.buf[i])
        end
        local inStringChecksum = ascii_hex_to_number(self.buf,
                                                     star_offset + 1, 2)
        if sum ~= inStringChecksum then
            -- KLUDGE Warn about bad checksum
            discard(self, end_offset + 1, "sentence with bad checksum")
            break
        end

        local fields = membuf:new(self.buf, star_offset)
        local timestamp = self.stamps[1][2]
        for i = 2, #self.stamps do
            if self.stamps[i][1] <= self.stream_pos + 1 then
                timestamp = self.stamps[i][2]
            else
                break
            end
        end

        -- get the "sentence type" field, and remove it
        local sentence_type_len = fields:search(0, ",", 1)
        local sentence_type = fields:extract(0, sentence_type_len)
        fields:remove(0, 1)  -- remove separator character

        out[#out + 1] = {
            protocol = "nmea",
            timestamp = timestamp,
            nmea = self.buf,
            nmea_len = end_offset + 1,
            sentence = sentence_type,
            fields = fields
        }

--[==[
dbg.debugf("NMEA [%g]: %s\n", timestamp, nmea)
for i = 1, #fields do
    dbg.debugf("    [%d]: %s\n", i, fields[i])
end
--]==]

        discard(self, end_offset + 1)
    end
    return out
end

return m
