--! PCRE (Perl-Compatible Regular Expressions) bindings for LuaJIT.
--! \file

--! \mainpage

require 'strict'

local ffi, pcre_ffi = unpack(require('pcre_ffi_defs'))
local C = ffi.C

local c_consts = {
    ovec_size = 30,
}

local pcre_defs = string.gsub([[

typedef struct {
    const char *regexp;
    pcre *pcre;
    pcre_extra *extra;
    const char *err[1];
    int err_offset[1];
    int ovec[$(ovec_size)];
} pcre_obj;

]], '%$%(([%w_]+)%)', function(s) return assert(c_consts[s]) end)

ffi.cdef(pcre_defs)

local function errorf(fmt, ...)
    error(string.format(fmt, ...))
end

local pcre_obj = {}

function pcre_obj:__gc()
    C.free(self.extra)
    C.free(self.pcre)
end

--! Match a string against the regular expression
--! \param input
--! \param captures
function pcre_obj:match(input, captures)
    local rc = pcre_ffi.pcre_exec(self.pcre, self.extra, input, #input,
        0, 0, self.ovec, c_consts.ovec_size)
        -- ffi.sizeof('pcre_obj.ovec') / ffi.sizeof('pcre_obj.ovec[0]'))
    if rc < 0 or not captures then
        return rc >= 0
    end
    if rc == 0 then
        error('Insufficient room for requested captures')
    end
    for i = 0, rc - 1 do
        captures[#captures + 1]
            = input:sub(self.ovec[i * 2] + 1, self.ovec[i * 2 + 1])
    end
    return rc
end

ffi.metatype('pcre_obj', { __index = pcre_obj })

--! It's a pcre class.
--? class pcre
local pcre = { }

--! Creates a new PCRE object
--! \param regexp 
function pcre:new(regexp, options)
    options = options or {}
    local pcre_compile_opts = 0
    local pcre_study_opts = 0
    local o = ffi.new('pcre_obj')
    o.regexp = ffi.cast('const char *', regexp)

    o.pcre = pcre_ffi.pcre_compile(o.regexp, pcre_compile_opts,
        o.err, o.err_offset, nil)
    if o.pcre == nil then
        errorf('pcre_compile() failed at offset %d: %s',
            o.err_offset[0], ffi.string(o.err[0]))
    end
    o.extra = pcre_ffi.pcre_study(o.pcre, pcre_study_opts, o.err)
    if o.err[0] ~= nil then
        errorf('pcre_study() failed: %s', ffi.string(o.err[0]))
    end
    return o
end

return pcre
