-- POSIX Thread Library 'libpthread'

local log = require 'log'
local ffi = require 'ffi'
local C = ffi.C

local PT = ffi.load('libpthread.so.0')

if ffi.os == 'Linux' then
    if ffi.arch == 'x64' then
        ffi.cdef[[ static const int __SIZEOF_PTHREAD_ATTR_T = 56; ]]
    else
        ffi.cdef[[ static const int __SIZEOF_PTHREAD_ATTR_T = 36; ]]
    end
end

ffi.cdef[[
typedef void *(*thread_func)(void *);

struct _pthread_fastlock {
    long int __status;
    int __spinlock;
};

typedef struct _pthread_descr_struct *_pthread_descr;

typedef struct {
    int __m_reserved;
    int __m_count;
    _pthread_descr __m_owner;
    int __m_kind;
    struct _pthread_fastlock __m_lock;
} pthread_mutex_t;

typedef struct {
    int __mutexkind;
} pthread_mutexattr_t;

int pthread_mutex_init(pthread_mutex_t *restrict mutex,
          const pthread_mutexattr_t *restrict attr);
int pthread_mutex_destroy(pthread_mutex_t *mutex);

int pthread_mutex_lock(pthread_mutex_t *mutex);
int pthread_mutex_unlock(pthread_mutex_t *mutex);

typedef uint64_t pthread_t;

void pthread_exit(void *retval);
pthread_t pthread_self(void);

enum {
    PTHREAD_CREATE_JOINABLE = 0,
    PTHREAD_CREATE_DETACHED = 1,
};

typedef union {
    int8_t __size[__SIZEOF_PTHREAD_ATTR_T];
    int64_t __align;
} pthread_attr_t;

int pthread_attr_init(pthread_attr_t *attr);

int pthread_attr_destroy(pthread_attr_t *attr);

int pthread_attr_setdetachstate(pthread_attr_t *attr, int detachstate);

int pthread_create(pthread_t *thread, const pthread_attr_t *attr,
                          void *(*start_routine) (void *), void *arg);

int pthread_join(pthread_t thread, void **retval);
]]

-- Lua C API

ffi.cdef[[
    static const int LUA_GLOBALSINDEX = -10002;
    typedef struct lua_State lua_State;
    lua_State *luaL_newstate(void);
    void luaL_openlibs(lua_State *L);
    void lua_close(lua_State *L);
    int luaL_loadfile(lua_State *L, const char *filename);
    int luaL_loadstring(lua_State *L, const char *s);
    int lua_pcall(lua_State *L, int nargs, int nresults, int errfunc);
    void lua_getfield(lua_State *L, int index, const char *k);
    ptrdiff_t lua_tointeger(lua_State *L, int index);
    void lua_settop(lua_State *L, int index);
]]

local pthread = {}

function pthread:mutex_init(mutex)
    PT.pthread_mutex_init(mutex, nil)
end

function pthread:lock(mutex)
    PT.pthread_mutex_lock(mutex)
end

function pthread:unlock(mutex)
    PT.pthread_mutex_unlock(mutex)
end

function pthread:exit()
    if not self.L then
        PT.pthread_exit(nil)
    end
end

function pthread:join()
    if self.L then
        if PT.pthread_join(self.tid[0], nil) ~= 0 then
            log:log(log.ERR,
                string.format("pthread:create(): pthread_join %s",
                              ffi.string(C.strerror(ffi.errno()))))
        end
    end
end

function pthread:create(code, data)
    self.tid = ffi.new('pthread_t[1]')
    self.code = ffi.new('char [?]', #code + 1)
    ffi.copy(self.code, code, #code)
    self.L = C.luaL_newstate()
    if self.L == nil then
        log:log(log.ERR, "pthread:create(): luaL_newstate error")
        return false
    end
    C.luaL_openlibs(self.L)
    if C.luaL_loadstring(self.L, self.code) ~= 0 then
        log:log(log.ERR, "pthread:create(): luaL_loadstring error")
        return false
    end
    if C.lua_pcall(self.L, 0, 1, 0) ~= 0 then
        log:log(log.ERR, "pthread:create(): lua_pcall error")
        return false
    end
    C.lua_getfield(self.L, C.LUA_GLOBALSINDEX, 'thread_func_ptr')
    local thread_func = C.lua_tointeger(self.L, -1);
    C.lua_settop(self.L, -2);
    if PT.pthread_create(self.tid, nil,
                             ffi.cast('thread_func', thread_func),
                             data) ~= 0 then
        log:log(log.ERR,
            string.format("pthread:create(): pthread_create %s",
                          ffi.string(C.strerror(ffi.errno()))))
        return false
    end
    return true
end

return pthread
