require 'middleclass'

local cjson = require "cjson"
local sqlite = require "sqlite_ffi"
local uuid = require "uuid"

local db_file = '/var/lib/armada/sessions.db'

local M = class('session_db')

function M:initialize(settings)
    self.db = sqlite:new(db_file)
    self.db:busy_timeout(1) -- KLUDGE this should be done differently!!!
    self:begin_transaction()
    self.db:exec([[
        CREATE TABLE IF NOT EXISTS sessions (
            session_key TEXT NOT NULL PRIMARY KEY,
            authorization_key TEXT NOT NULL,
            timestamp FLOAT NOT NULL DEFAULT 0,
            user_id TEXT NOT NULL DEFAULT "",
            state TEXT NOT NULL
        );
    ]])
    self:commit_transaction()
    self.read_session_stmt = self.db:prepare([[
        SELECT * FROM sessions WHERE session_key = ?;
    ]])
    self.get_authorized_session_stmt = self.db:prepare([[
        SELECT *
        FROM sessions
        WHERE authorization_key = ?
        ORDER BY timestamp DESC
        LIMIT 1;
    ]])
    self.write_session_stmt = self.db:prepare([[
        INSERT OR REPLACE INTO sessions (
            session_key,
            authorization_key,
            timestamp,
            user_id,
            state
        ) VALUES (
            :session_key,
            :authorization_key,
            :timestamp,
            :user_id,
            :state
        );
    ]])
    self.end_session_stmt = self.db:prepare([[
        DELETE FROM sessions WHERE session_key = ?;
    ]])
end

function M:begin_transaction()
    self.db:exec([[
        BEGIN TRANSACTION;
    ]])
end

function M:commit_transaction()
    self.db:exec([[
        COMMIT TRANSACTION;
    ]])
end

function M:update_session(s)
    local session = {
        session_key = s.session_key or uuid.generate(),
        authorization_key = s.authorization_key, -- KLUDGE? or uuid.generate(),
        timestamp = s.timestamp or 0, -- KLUDGE
        user_id = s.user_id,
        state = cjson.encode(s.state),
    }
    self:begin_transaction()
    self.write_session_stmt:exec(session)
    self:commit_transaction()
    return session
end

local function reform_session(s)
    if type(s) == 'table' and #s == 1 then
        s = s[1]
        if s and s.state then
            s.state = cjson.decode(s.state)
        end
    else
        s = nil
    end
    return s
end

function M:get_session(session_key)
    self:begin_transaction()
    local r = self.read_session_stmt:exec{session_key}
    self:commit_transaction()
    return reform_session(r)
end

function M:get_session_from_authorization(authorization_key)
    self:begin_transaction()
    local r = self.get_authorized_session_stmt:exec{authorization_key}
    self:commit_transaction()
    return reform_session(r)
end

function M:end_session(session_key)
    self:begin_transaction()
    local r = self.end_session_stmt:exec{session_key}
    self:commit_transaction()
    return r
end

function M:get_all_sessions()
    return self.db:exec[[
        SELECT * FROM sessions;
    ]]
end

return M
