-- Proxy node support functions used by the test engine
local ev = require 'ev'
local network_db = require 'network_db'
local log = require 'log'
local proxy_node = require 'test_engine.proxy_node'
local gps_node = require 'test_engine.gps_node'
local can_node = require 'test_engine.can_node'
local digital_out_node = require 'test_engine.digital_out_node'
local ffi = require 'ffi'

local m = {}

local function errorf(fmt, ...)
    error(string.format(fmt, ...))
end

-- Function used to create a proxy node.
function m:create_proxy_node(md)
    if not md.connector then
        errorf('Input channel %s does not have a connector', md.name)
        return
    end
    if not md.data_type then
        errorf('Input channel %s does not have a data type', md.name)
        return
    end

    local orig_host = md.connector:match('^.*%@(.+)')
    local net_db = network_db:new()
    local host, port, ip = net_db:resolve_service(orig_host)
    net_db:close()
    if not host or not port or not ip then
        errorf("Proxy for channel %s not found", md.name)
        return
    end

    return proxy_node:new(ev.Loop.default, md, host, port, ip,
        self.summary.pipe_rate or 5)
end

function m:create_digital_out_node(md)
    if not md.connector then
        errorf('Input channel %s does not have a connector', md.name)
        return
    end

    local orig_host = md.connector:match('^.*%@(.+)')
    local net_db = network_db:new()
    local host, port, ip = net_db:resolve_service(orig_host)
    net_db:close()
    if not host or not port or not ip then
        errorf("Proxy for channel %s not found", md.name)
        return
    end

    return digital_out_node:new(ev.Loop.default, md, host, port, ip)
end

function m:create_gps_signal_node(md, raw_pn, pipe_rate)
    local gn = gps_node.new(md, pipe_rate or 5, raw_pn)
    if gn == nil then
        errorf("Unable to create GPS signal node for %s", md.name)
    end
    if raw_pn.gs_list == nil then
        raw_pn.gs_list = gn
    else
        local p = raw_pn.gs_list
        while p.next ~= nil do p = p.next end
        p.next = gn
    end
    return gn
end

function m:create_mx878_proxy_node(md, pipe_rate)
    local orig_host = md.connector:match('^.*%@(.+)')
    local net_db = network_db:new()
    local host, port, ip = net_db:resolve_service(orig_host)
    net_db:close()
    if not host or not port or not ip then
        error(string.format("Proxy for channel '%s' not found", md.name))
    end
    local pn, start = proxy_node:new(ev.Loop.default, md, host, port, ip,
        pipe_rate or 5)
    return pn, start
end

function m:create_mx_can_node(md, can_connector)
    local pn, start, cn, set_time

    if not md.data_type then
        errorf('Input channel %s does not have a data type', md.name)
        return
    end

    -- If a connection to the proxy does not exist for this connector, create
    -- one.
    if not self.mx_can_channels[can_connector] then
        local orig_host = md.connector:match('^.*%@(.+)')
        local net_db = network_db:new()
        local host, port, ip = net_db:resolve_service(orig_host)
        net_db:close()
        if not host or not port or not ip then
            errorf("Proxy for channel %s not found", md.name)
            return
        end

        pn, start, set_time = proxy_node:new(ev.Loop.default, md, host, port,
            ip, self.summary.pipe_rate or 5)

        self.mx_can_channels[can_connector] = pn
    end

    if not self.mx_can_channels[can_connector] then
        errorf("No MX proxy node for channel %s", md.name)
        return
    end

    -- Generate the reference string for this channel
    -- NOTE:  This assumes the CAN signals are allocated on the MX in test
    -- order!
    local num = 0
    local cs = self.mx_can_channels[can_connector].cs_list
    while cs ~= nil do
        num = num + 1
        cs = cs.next
    end
    md._reference = string.format('CanReceiver_Connector%s_Decoder%d.Signal1',
        md.connector:match('@can(%d+)'), num + 1)

    -- Create a new MX CAN signal node
    local cn = can_node.new(md, self.summary.pipe_rate or 5)
    if cn == nil then
        errorf("Unable to create CAN signal node for %s", md.name)
    end
    cn.pn = ffi.cast('void *', self.mx_can_channels[can_connector])

    -- Attach this CAN signal node to the MX connector
    if self.mx_can_channels[can_connector].cs_list == nil then
        self.mx_can_channels[can_connector].cs_list = cn

    else
        local p = self.mx_can_channels[can_connector].cs_list
        while p.next ~= nil do p = p.next end
        p.next = cn
    end

    return pn or nil, start or nil, cn, set_time
end

return m
